from os.path import basename
from IPython.display import display, HTML
import numpy as np
import pandas as pd
import sys
from .base import date_range_array, relevant_files_list, files_containing_str


def show_mem_usage():
    """
    Displays memory usage from inspection
    of global variables in this notebook

    taken from
    http://practicalpython.blogspot.com/2017/03/monitoring-memory-usage-in-jupyter.html
    """
    gl = sys._getframe(1).f_globals
    vars = {}
    for k, v in list(gl.items()):
        # for pandas dataframes
        if hasattr(v, 'memory_usage'):
            mem = v.memory_usage(deep=True)
            if not np.isscalar(mem):
                mem = mem.sum()
            vars.setdefault(id(v), [mem]).append(k)
        # work around for a bug
        elif isinstance(v, pd.Panel):
            v = v.values
        vars.setdefault(id(v), [sys.getsizeof(v)]).append(k)
    total = 0
    for k, (value, *names) in vars.items():
        if value > 1e6:
            print(names, "%.3fMB" % (value / 1e6))
        total += value
    print("%.3fMB" % (total / 1e6))


def disp(df):
    """
    display a pandas DataFrame without the index
     (wrapper of IPython.display.display)
    designed for use within Jupyter notebook
    """
    display(HTML(df.to_html(index=False)))


def notebook_contains(search_str='',
                      on_docker=True,
                      git_dir='~/git/experiments/',
                      start_date='2015-01-01', end_date='2018-12-31',
                      exclude_str='checkpoint',
                      include_prefix=False,
                      prefix='notebooks/'):
    """
    return all jupyter notebooks containing search_str in specified time window.
    TO DO: extend to list of search_strings, extend to list of exclude_strings.
    """
    if on_docker:
        base_dir = "/home/jovyan/work/"
    else:
        base_dir = git_dir[:]
    dates = date_range_array(start=start_date, end=end_date)
    rel_files = relevant_files_list(base_dir, dates, exclude_str)
    files = files_containing_str(search_str, rel_files)
    if prefix[-1] == '/':
        prefix = prefix[:-1]
    if include_prefix:
        return [prefix+el.split(basename(prefix))[-1] for el in files]
    else:
        return [el.split(basename(prefix))[-1] for el in files]


from matplotlib import pyplot as plt
import seaborn as sns
import pandas as pd
from ..ml.linear_algebra import distmat


def scatter_2d(orig_df: pd.DataFrame, colx, coly, label_col,
               xmin=None, xmax=None, ymin=None, ymax=None):
    """
    Return scatter plot of 2 columns in a DataFrame, taking labels as colours.
    """
    plt.scatter(orig_df[colx], orig_df[coly],
                c=orig_df[label_col].values, cmap='viridis')
    plt.colorbar()
    plt.xlim(xmin, xmax)
    plt.ylim(ymin, ymax)
    plt.show()


def visualise_dist(fframe=None, metric='euclidean'):
    """
    Plot a distance matrix from a DataFrame containing only feature columns.
        The plot is a heatmap, and the distance metric is specified with `metric`
    """

    plt.figure(figsize=(14, 10))
#     ax = plt.gca()
    sns.heatmap(distmat(fframe, metric=metric))
    plt.show()


import numpy as np


def matlab3d_style(np3_array):
    """
    Return MATLAB style display of 3d Numpy array
    :param np3_array: np.array with shape (x,y,3)
    :return:
    """
    a = np3_array.copy()
    print(np.array([a[:, :, 0], a[:, :, 1], a[:, :, 2]]))

    return np3_array


def identity(x):
    return x


def num_cols(arr: np.array):
    """
    returns number of columns in numpy array
    Note, this work around is necessary as the shape tuple for a one column array
        returns (x, ), so arr.shape[1] doesn't always work.
    """
    return arr[:, np.newaxis].shape[-1]



from scipy.spatial import distance
from sklearn.metrics.pairwise import cosine_similarity


def cos_sim_sk(a, b):
    """
    Wrapper around sklearn.metrics.pairwise.cosine_similarity
    Useful for bypassing DeprecationWarning by reshaping.

    Inputs are of type numpy.ndarray
    """
    return cosine_similarity(a.reshape(1, -1), b.reshape(1, -1))[0][0]


def cos_sim_sci(vec1, vec2):
    """
    Returns cosine similarity of 2 vectors.
    Uses scipy.spatial.distance.cosine
    """
    return 1 - distance.cosine(vec1, vec2)


def distmat(fframe=None, metric=None):
    """
    Generate a distance matrix from a DataFrame containing only feature columns.
        The distance metric is specified with `metric`.
    If called with no args, return list of possible metrics.

    """
    if fframe is None:
        d_metric = ["braycurtis", "canberra", "chebyshev", "cityblock", "correlation",
                    "cosine", "dice", "euclidean", "hamming", "jaccard", "kulsinski",
                    "mahalanobis", "matching", "minkowski", "rogerstanimoto",
                    "russellrao", "seuclidean", "sokalmichener", "sokalsneath",
                    "sqeuclidean", "wminkowski", "yule"]
        return d_metric
    return distance.cdist(fframe.as_matrix(), fframe.as_matrix(), metric)


import zipfile
import boto3
import io
import datetime
import pandas as pd


def get_creds(cred_fpath=None):
    """helper function to obtain aws keys
    """
    if cred_fpath is not None:
        print("reading keys from credentials file")
        keys = pd.read_csv(cred_fpath, sep="=")
        myAccessKey = keys.loc['aws_access_key_id ']['[default]'].strip()
        mySecretKey = keys.loc['aws_secret_access_key ']['[default]'].strip()
    else:
        print("type in aws access key yo:")
        myAccessKey = input()
        print("type in aws secret key yo:")
        mySecretKey = input()
    return myAccessKey, mySecretKey


def s3zip_func(s3zip_path, _func, include: list = [], exclude: list = [], cred_fpath=None, verbose=False, **kwargs):
    """
    unzip a zip file on s3 and perform func with kwargs.
     func must accept `fpath` and `fname` as key word arguments.
    fpath: pointer to unzipped subfile in zip file
    fname: str of subfile in zip file

    adapted from https://stackoverflow.com/questions/23376816/python-s3-download-zip-file
    """
    s3bucket, s3zip = s3zip_path.split("s3://")[-1].split('/', 1)
    myAccessKey, mySecretKey = get_creds(cred_fpath=cred_fpath)

    session = boto3.session.Session(
        aws_access_key_id=myAccessKey,
        aws_secret_access_key=mySecretKey
    )

    s3 = session.resource("s3")
    bucket = s3.Bucket(s3bucket)
    obj = bucket.Object(s3zip)

    results = []

    with io.BytesIO(obj.get()["Body"].read()) as tf:

        # rewind the file
        tf.seek(0)

        # Read the file as a zipfile and process the members
        with zipfile.ZipFile(tf, mode='r') as zipf:
            for subfile in zipf.namelist():
                if verbose:
                    print("current time is {}".format(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')))
                if include:
                    if subfile in include:
                        if verbose:
                            print("{} opened.".format(subfile))
                        result = _func(fpath=zipf.open(subfile), fname=subfile, **kwargs)
                        results.append((subfile, result))
                else:
                    if subfile in exclude:
                        if verbose:
                            print("{} skipped.".format(subfile))
                    else:
                        if verbose:
                            print("{} opened.".format(subfile))
                        result = _func(fpath=zipf.open(subfile), fname=subfile, **kwargs)
                        results.append((subfile, result))
    return results



"""
helper functions to use with pandas library.
Conventions:
df -> pd.DataFrame

Ideology:
- Write functions that do one thing
- Try write functions that are idempotent
"""

from functools import reduce
from collections import namedtuple
import numpy as np
import pandas as pd
import re


def drop_uniq_cols(df):
    """
    Returns DataFrame with columns that have more than one value (not unique)
    """
    return df.loc[:, df.apply(pd.Series.nunique) > 1]


def non_x_cols(df, x):
    """
    Returns DataFrame with columns that contain values other than x.
    Drops any column that only contains x.
    """
    cond_1 = (df.apply(pd.Series.nunique) == 1)
    cond_2 = (df.iloc[0] == x)
    return df.loc[:, ~(cond_1 & cond_2)]


def df_duplicate_row(df, row, row_cols):
    """
    (Check logic)
    Creates numpy array from df (pd.DataFrame),
    returns non_duplicate rows.
    """
    X = df[row_cols].as_matrix().astype(np.float)
    fv_ix = [ix for ix, el in enumerate(X) if (el == row.values).all()]
    df_rows = df.ix[fv_ix]
    return df_rows


def filt(orig_df, **params):
    """
    Filter DataFrame on any number of equality conditions.
    Example usage:
    >> filt(df,
               season="summer",
               age=(">", 18),
               sport=("isin", ["Basketball", "Soccer"]),
               name=("contains", "Armstrong")
               )
    >> a = { 'season': "summer", 'age': (">", 18)}
    >> filt(df, **a) # can feed in dict with **dict notation
    notes:
        any input with single value is assumed to use "equivalent" operation and is modified.
        numpy.all is used to apply AND operation element-wise across 0th axis.
        NA values are filled to False for conditions.
    """
    input_df = orig_df.copy()

    if not params.items():
        return input_df

    def equivalent(a, b): return a == b

    def greater_than(a, b): return a > b

    def greater_or_equal(a, b): return a >= b

    def less_than(a, b): return a < b

    def less_or_equal(a, b): return a <= b

    def isin(a, b): return a.isin(b)

    def notin(a, b): return ~(a.isin(b))

    def contains(a, b): return a.str.contains(b)

    def notcontains(a, b): return ~(a.str.contains(b))

    def not_equivalent(a, b): return a != b

    operation = {"==": equivalent,
                 "!=": not_equivalent,
                 ">": greater_than,
                 ">=": greater_or_equal,
                 "<": less_than,
                 "<=": less_or_equal,
                 "isin": isin,
                 "notin": notin,
                 "contains": contains,
                 "notcontains": notcontains}

    cond = namedtuple('cond', 'key operator val')

    filt_on = [(el[0], el[1]) if isinstance(el[1], tuple) else (el[0], ("==", el[1]))
               for el in params.items()]  # enforcing equivalence operation on single vals.
    conds = [cond(el[0], el[1][0], el[1][1]) for el in filt_on]
    logic = [operation[x.operator](input_df[x.key], x.val).fillna(False) for x in conds]
    return input_df[np.all(logic, axis=0)]


def filt_read_csv(file_path, chunksize=100000, **filt_conds):
    """
    read a csv in chunks while filtering on conditions
    wrapper of pd.read_csv and uses pandas_.base.filt
    """
    iter_csv = pd.read_csv(file_path, iterator=True, chunksize=chunksize)
    return pd.concat([filt(chunk, **filt_conds) for chunk in iter_csv], ignore_index=True)


def display_max_rows(max_num=500):
    """sets number of rows to display (useful in Jupyter environment)
    """
    pd.set_option('display.max_rows', max_num)


def assign_index(input_df, index_col='idx'):
    """Return input DataFrame with id col.
    """
    out_df = input_df.reset_index(drop=True)
    out_df.loc[:, index_col] = out_df.index
    return out_df


def enumerate_col(input_df, enumerate_over, new_col="newCol"):
    """
    Enumerate unique entries in enumerate_over_col and add new column newCol.
    """
    new_df = input_df.copy()
    unique_vals = new_df[enumerate_over].unique()
    enumerated_dict = {vals: ix for ix, vals in enumerate(unique_vals)}
    new_df.loc[:, new_col] = new_df[enumerate_over].map(lambda x: enumerated_dict[x])
    return new_df


def concat_df_list(list_csv, comment='#', header='infer', sep=',', nrows=None):
    """Concatenates list of dataframes first created by list comprehension.

    """
    return pd.concat([pd.read_csv(el, comment=comment, header=header, sep=sep, nrows=nrows)
                      for el in list_csv], ignore_index=True)


def merge_dfs(dataframe_list, join_meth='outer', on=None, left_index=False, right_index=False):
    """
    Merge together a list of DataFrames.
    """

    def simple_merge(df1, df2):
        return pd.merge(left=df1, right=df2, on=on,
                        left_index=left_index, right_index=right_index, how=join_meth)

    merged_frame = reduce(simple_merge, dataframe_list)
    return merged_frame


def uniq_x_y(X, y, output_index=False):
    """
    Given a X,y pair of numpy arrays (X:2D, y:1D),
    return unique rows (arrays) of X, with corresponding y values.
    """
    uniq_df = pd.DataFrame(X).drop_duplicates()
    X_uniq = uniq_df.values
    y_uniq = y[uniq_df.index]
    if output_index:
        return X_uniq, y_uniq, uniq_df.index
    return X_uniq, y_uniq


def gby_count(df, col, col_full_name=False):
    """
    Perform groupby count aggregation, rename column to count,
    and keep flat structure of DataFrame.
    Wrapper on pd.DataFrame.groupby
    """
    if col_full_name:
        col_name = "count_" + col
    else:
        col_name = "count"
    return df.groupby(col).size().reset_index(name=col_name).sort_values(by=col_name, ascending=False)


def distribute_rows(df, key_col):
    """
    Distributes rows according to the series found in key_col.
    Example:
        >> df = pd.DataFrame([(el+3, el) for el in range(4)], columns=['x', 'n'])
        >> df.T
           0  1  2  3
        x  3  4  5  6
        n  0  1  2  3
        >> df.loc[np.repeat(df.index.values,df.n)].T
           1  2  2  3  3  3
        x  4  5  5  6  6  6
        n  1  2  2  3  3  3
    """
    distribution_index = np.repeat(df.index.values, df[key_col])
    return df.loc[distribution_index]


def resample_by_col(df, gby_col='Cluster ID', scale_by='identity'):
    """
    Resample the rows of DataFrame.
    Groupby is first performed on gby_col, then counts are rescaled by
    function defined in scale_by (identity, log, sqrt).
    Groupby DataFrame is then resampled by new distribution, and an outer
    join is performed with original.drop_duplicates(gby_col).

    Note: inputting `df` and `df.drop_duplicates(gby_col)` lead to different
    results in `gby_col_count` col. Always input original.

    Improvements: create all_n func and do pass function curried like in scala
    """
    gby_df = df.groupby(gby_col).size().reset_index(name=gby_col + "_count")

    def identity(x): return x

    def all_one(x): return 1

    def all_two(x): return 2

    def all_three(x): return 3

    scale_funcs = {'log': np.log, 'sqrt': np.sqrt, 'identity': identity,
                   'all_one': all_one, 'all_two': all_two, 'all_three': all_three}
    distribution = gby_df[gby_col + "_count"].map(scale_funcs[scale_by]).map(np.ceil).map(np.int)
    df_new_dist = gby_df.loc[np.repeat(gby_df.index.values, distribution)]
    return pd.merge(df_new_dist, df.drop_duplicates(gby_col), on=gby_col, how='outer')


def convert_all_elements_tuple(input_df):
    """
    If all elements are lists or numpy arrays, these are not hashable.
    So we can convert to tuples, in order to be able to perform groupbys.
    """
    return input_df.applymap(tuple)


def explode_rows_vertically(df):
    """
    For use on DataFrames containing lists in each field.
    Note: per row, lists in each column must be equal in length.
        They can, however, differ in length from row to row.
    Explode lists in each column of input DataFrame vertically,
        one row at a time.
    Return DataFrame of exploded elements.
    Note we could also use `+= zip(*[df[col][ix] for col in df])`
    """
    exploded_list = []
    for ix, row in df.iterrows():
        exploded_list += zip(*row.values)
    return pd.DataFrame(exploded_list)


def str_list_to_tuple_str_series(col, regex_pattern='[A-Z]\d+'):
    """
    Convert string of lists into tuples of strings,
    for each row in column.
    regex_pattern determines string tokens.
    """
    if not isinstance(col[0], str):
        print("error: str expected, instead {} found.".format(type(col[0])))
        return col
    else:
        p = re.compile(regex_pattern)
        return col.apply(p.findall).apply(tuple)


def df_to_namedtups(df, name="Pandas"):
    """
    Given a DataFrame, df, return a list of namedtuples
    """
    return list(df.itertuples(index=False, name=name))


def gby_uniqlists(in_df, pivot_col=None, col=None) -> pd.DataFrame:
    """
    Computes aggregates for each unique value in `pivot_col` and returns
        new DataFrame with list of unique values in col + nunique in col.

    :param in_df: pd.DataFrame; input DataFrame
    :param pivot_col: str; column to groupby (pivot) off
    :param col: str; column to aggregate values on.
    :return: pd.DataFrame; groupby aggregates DataFrame with unique vals in list + length.
    """

    gby_temp = in_df.groupby(pivot_col)[col].apply(set).apply(list).reset_index()
    return gby_temp.assign(len_=gby_temp[col].apply(len)).rename(
        columns={'len_': col + '_uniq_len', col: col + '_uniq'})


def gby_multiframe(in_df, pivot_col=None, cols=None) -> pd.DataFrame:
    """
    wrapper of velpy.helper.pandas_.gby_uniqlists performed over list of cols,
        then joined on pivot_col.
    :param in_df: pd.DataFrame; input DataFrame
    :param pivot_col: str; column to groupby (pivot) off
    :param cols: list[str]; list of columns to aggregate values on.
    :return: pd.DataFrame; merged groupby aggregates DataFrame
    """
    cols = [col for col in cols if col in in_df.columns]
    return merge_dfs([gby_uniqlists(in_df, pivot_col=pivot_col, col=col)
                      for col in cols], on=pivot_col)


def nonuniq(orig_df):
    """
    return DataFrame of sorted non-uniq counts of fields (greater than 1).
    """
    return (orig_df.apply(pd.Series.nunique)
            .sort_values(ascending=False)
            .reset_index(name='uniq_count')
            .pipe(lambda x: x[x['uniq_count'] > 1])
            )


def update_sub(orig_df, pivot_col=None, pivot_val=None, new_col=None, new_val=None):
    """
    update a subset of DataFrame based on fixed boolean matching.
    """
    out_df = orig_df.copy()
    out_df.loc[out_df[pivot_col] == pivot_val, new_col] = new_val
    return out_df


def sorted_cols(orig_df):
    """
    sort columns by name and return df
    """
    cols = sorted(orig_df.columns)
    return orig_df[cols]


def ddupe_with_precision(orig_df, precision=10):
    """
    round original DataFrame to a given precision, drop_duplicates,
        return original dataframe with those rows indexed.
    """
    return orig_df.loc[orig_df.round(precision).drop_duplicates().index]


def mapping_2cols(orig_df, colx, coly):
    """
    return dict mapping vals in colx to coly. Order is important.
    """
    return dict(orig_df[[colx, coly]].drop_duplicates().values)


def newcol_mapped(orig_df, orig_col, new_col, mapping):
    """
    map a new column from an old column and a mapping dictionary
    """
    new_vals = orig_df[orig_col].map(mapping).values
    return orig_df.assign(**{new_col: new_vals})


def prepend_comment(orig_df):
    """
    prepend the first column name with a '#'
    """
    input_df = orig_df.copy()
    first_col = input_df.columns[0]
    return input_df.rename(columns={first_col: '#' + first_col})


def remove_prepended_comment(orig_df):
    """
    remove the '#' prepend to the first column name
    Note: if there is no comment, this won't do anything (idempotency).
    """
    input_df = orig_df.copy()
    first_col = input_df.columns[0]
    return input_df.rename(columns={first_col: first_col.replace('#', "")})


def remove_double_quotes(orig_df: pd.DataFrame, quote_cols, all_cols=False) -> pd.DataFrame:
    """
    Replace double quotes found in fields with two single quotes.
     This must be done for SQL queries to correctly parse fields in Hive (and others)
    """

    def replace_quotes(x):
        if isinstance(x, str):
            return x.replace("\"", "''")
        return x

    out_df = orig_df.copy()
    if all_cols:
        return out_df.applymap(replace_quotes)
    else:
        out_df[quote_cols] = out_df[quote_cols].applymap(replace_quotes)
        return out_df


def apply_uniq(df, orig_col, new_col, _func):
    """
    Apply func to only unique entries and join with original to fill.
    Answered for:
    https://stackoverflow.com/questions/46798532/how-do-you-effectively-use-pd-dataframe-apply-on-rows-with-duplicate-values/

    """
    out_df = df.copy()
    if new_col in out_df:
        out_df = out_df.drop(new_col, axis='columns')
    return out_df.merge(out_df[[orig_col]]
                        .drop_duplicates()
                        .assign(**{new_col: lambda x: x[orig_col].apply(_func)}
                                ), how='inner', on=orig_col)


def mult_types(df):
    """
    Return series of columns with unique number of types per col

    Can be useful to see which cols have more than 1 type, for example:
    >> counts = mult_types(df)
    >> counts[counts != 1]
    """
    return df.applymap(type).drop_duplicates().apply(pd.Series.nunique)


def collapse_hierarchy(orig_df: pd.DataFrame):
    """
    this collapses multilevel pandas columns into a single level by joining tokens with underscore.
    Example:
        orig_cols = [('a', 'max'), ('a', 'size'), ('b', '')]
         new_cols = ['a_max', 'a_size', 'b']
    """
    gby = orig_df.copy()
    if isinstance(gby.columns, pd.core.indexes.multi.MultiIndex):
        gby.columns = ['_'.join(col).rstrip('_') for col in gby.columns.values]
    else:
        print("columns not MultiIndex, returning original.")
    return gby




from ..s3_ import get_creds


def s3spark_init(cred_fpath="~/.aws/credentials"):
    """
    initialise SparkSession for use with Jupyter and s3 SQL queries
    Returns spark session

    aws credentials file default path: "~/.aws/credentials"
    """
    import os
    os.environ['PYSPARK_SUBMIT_ARGS'] = '--packages com.amazonaws:aws-java-sdk:1.10.34,org.apache.hadoop:hadoop-aws:2.6.0 pyspark-shell'

    from pyspark.sql import SparkSession

    spark = SparkSession \
        .builder \
        .appName("using_s3") \
        .getOrCreate()

    hadoopConf = spark.sparkContext._jsc.hadoopConfiguration()
    myAccessKey, mySecretKey = get_creds(cred_fpath=cred_fpath)

    hadoopConf.set("fs.s3.impl", "org.apache.hadoop.fs.s3native.NativeS3FileSystem")
    hadoopConf.set("fs.s3.awsAccessKeyId", myAccessKey)
    hadoopConf.set("fs.s3.awsSecretAccessKey", mySecretKey)
    return spark


def register_sql(spark, files, schema=None, sep=None, table_name="table", return_count=False):
    """
    Register a list of files as a SQL temporary view.

    parameters:
    - files is overloaded: can be one file path or list of file paths.
    - spark: pyspark.sql.SparkSession
    - table_name: this is how we will refer to table in SQL query
    Schema of files must be the same for the table

    Example usage after registering table:
    >> DF = spark.sql("SELECT * FROM table")
    """
    tempFiles = spark.read.csv(files, schema=schema, sep=sep, header=True)
    tempFiles.createOrReplaceTempView(table_name)
    if return_count:
        return spark.sql("SELECT * FROM {}".format(table_name)).count()



"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='seipy',
    version='1.3.0',
    description='Helper functions for data science',
    long_description=long_description,
    url='https://github.com/Seiji-Armstrong/seipy',
    author_email='seiji.armstrong@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='pandas numpy spark jupyter data-science machine-learning s3',
    packages=find_packages(exclude=[]),
    install_requires=['pandas',
                      'scipy',
                      'scikit-learn',
                      'numpy',
                      'pandas',
                      'matplotlib',
                      'scapy-python3',
                      'seaborn',
                      'ipython',
                      'boto3',
                      'pyspark'],
    python_requires='>=3',
)



import subprocess
from os import walk
from os.path import join
from functools import reduce
from collections import Counter, namedtuple
import ast
import mmap
import re
import json
import pandas as pd


class Stack:
    """
    simple stack class
    """
    def __init__(self):
        self.items = []

    def isEmpty(self):
        return self.items == []

    def push(self, item):
        self.items.insert(0, item)

    def pop(self):
        return self.items.pop(0)

    def peek(self):
        return self.items[0]

    def size(self):
        return len(self.items)


class LabeledData(dict):
    """Container object for datasets

    Dictionary-like object that exposes its keys as attributes.

    (copied from sklearn.datasets.base.Bunch)
    """

    def __init__(self, **kwargs):
        # type: (object) -> object
        super(LabeledData, self).__init__(kwargs)

    def __setattr__(self, key, value):
        self[key] = value

    def __dir__(self):
        return self.keys()

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)

    def __setstate__(self, state):
        # https://github.com/scikit-learn/scikit-learn/issues/6196
        pass


def issue_shell_command(cmd: str, my_env=None):
    """
    Issues a command in a shell and returns the result as str.

    Parameters:
    cmd - command to be issued (str)

    In python3.x, stdout,stderr are both b'' (byte string literal: bytes object)
        and must be decoded to UTF-8 for string concatenation etc
    Example usage (simple):
    >> issue_shell_command(cmd="ls")
    Example usage (more involved):
    >> s3dir = "s3://..."; issue_shell_command("aws s3 ls --recursive {}".format(s3dir))
    """
    pipe = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE, env=my_env)

    return pipe.stdout.strip().decode('UTF-8') + '\n' + pipe.stderr.strip().decode('UTF-8')


def merge_dicts(*dict_args):
    """
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    """
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def merge_two_dicts(dict_1, dict_2):
    """
    Given two dicts, return one merged dict.
    """
    return {**dict_1, **dict_2}


def merge_multi_dicts(*dict_args):
    """
    Given multiple dicts, return one merge dict by calling `reduce` on `merge_two_dicts`.
    ALternative: reduce(lambda d1,d2: {**d1,**d2}, dict_args[0])
    """
    return reduce(merge_two_dicts, dict_args[0])


def swap_key_val_dict(a_dict):
    """
    swap the keys and values of a dict
    """
    return {val: key for key, val in a_dict.items()}


def enumerate_with_prefix(a_list, prefix='pre_'):
    """
    given a list, return a list enumerated with prefix.
    """
    num_digits = len(str(len(a_list)))  # eg 5 -> 1, 15 -> 2, 150 -> 3 etc.
    enum_list = [prefix + str(idx).zfill(num_digits)
                 for idx, el in enumerate(a_list)]
    return enum_list


def relevant_files(root_dir, include_regex='', exclude="*****"):
    """Return list of files with inclusion regex and exclusion regex.

    inputs:
    "root_dir" is the root directory
    "include_regex" is the string that is searched for within filenames
    "exclude_regex" is the string that will exclude files if found in name.

    returns:
    list of filenames in all subsequent folders, matched with include_regex.
    """
    f_list = []
    for (dir_path, _, file_names) in walk(root_dir):
        f_list.extend(join(dir_path, filename)
                      for filename in file_names
                      if (include_regex in filename))
    f_list = [el for el in f_list if exclude not in el]
    return f_list


def relevant_files_list(root_dir, include_list=[], exclude="*****"):
    """Return list of files with inclusion regex in list and exclusion regex.

    inputs:
    "root_dir" is the root directory
    "include_list" is list of strings that is searched for within filenames
    "exclude_regex" is the string that will exclude files if found in name.

    returns:
    list of filenames in all subsequent folders, matched with include_regex.
    """
    f_list = []
    for (dir_path, _, file_names) in walk(root_dir):
        f_list.extend(join(dir_path, filename)
                      for filename in file_names
                      if any(el in filename for el in include_list))
    f_list = [el for el in f_list if exclude not in el]
    return f_list


def files_containing_str(str_, file_list):
    """
    return all files in provided file_list containing str_.
    """
    return [el for el in file_list if str_in_file(str_, el)]


def date_range_array(start='2016-12-1', end='2016-12-5', strf='%Y-%m-%d'):
    """
    Return a numpy.array of dates in range between start and stop.
    Format of date string is given by strf.
    Wrapper around pandas.date_range
    """
    return pd.date_range(start=start, end=end).map(lambda x: x.strftime(strf))


def str_in_file(str_, file_):
    """
    Returns Boolean: True if str_ in file_, otherwise False.
    First converts string to Bytes.
    Uses mmap in order not to read entire file each time.
    """
    b_ = str_.encode()
    with open(file_, 'rb', 0) as file, \
            mmap.mmap(file.fileno(), 0, access=mmap.ACCESS_READ) as s:
        if re.search(b_, s):
            return True
        return False


def str_from_collec(collection, no_space=False):
    """
    Given a collection, return a string of values
        Collection can be: column of DataFrame (Series), list, etc.

    Useful in creating fv_str strings for example, by feeding row of DataFrame
    containing only feature columns
    """
    if no_space:
        return ','.join(map(str, collection))
    else:
        return ', '.join(map(str, collection))


def uniq_tokens_in_nested_col(col_series):
    """
    Given a column in a dataframe containing lists of tokens,
        return unique tokens.
    Can also receive a list of lists, pd.Series of lists, np.array of lists.
    """
    return set([el for sublist in col_series for el in sublist])


def uniq_tokens_css(thresh, *collection_vals):
    """
    Extract unique tokens in comma separated strings, contained in a collection,
        return list of unique tokens that occur more than a threshold value.
    Apply sorted to output for reproducible output.
    """
    temp_lists = []
    for vals in collection_vals:
        temp_lists += [el.split(',') for el in vals]
    temp_fulllist = [el.strip() for sublist in temp_lists for el in sublist]
    temp_count = Counter(temp_fulllist)
    return sorted([k for k, v in list(temp_count.items()) if v > thresh])


def str_list_to_list_str(str_list, regex_pattern='[A-Z]\d+'):
    """
    Turn a string of a list into a list of string tokens.
    Tokens determined by regex_pattern
    """
    p = re.compile(regex_pattern)
    return p.findall(str_list)


def save_json(dict_to_json, save_path):
    """
    Saves dict to json file
    """
    out_file = open(save_path, "w")
    # save the dictionary to this file
    json.dump(dict_to_json, out_file, indent=4)
    # close the file
    out_file.close()
    print("json file saved at: {}".format(save_path))


def load_json(path_to_load):
    """
    Loads json from file
    """
    with open(path_to_load, 'r') as f:
        return json.load(f)


def vals_sortby_key(dict_to_sort):
    """
    sort dict by keys alphanumerically, then return vals.
    Keys should be "feat_00, feat_01", or "stage_00, stage_01" etc.
    """
    return [val for (key, val) in sorted(dict_to_sort.items())]


def diff_2_lists(list_1, list_2):
    """
    return elements not common to both lists
    Symmetric difference operator ^ is used instead of -.
    """
    return list(set(list_1) ^ set(list_2))


def file_len(fname):
    """
    count number of lines in file
    copy-pasted from http://stackoverflow.com/questions/845058/how-to-get-line-count-cheaply-in-python
    """
    with open(fname) as f:
        for i, l in enumerate(f):
            pass
    return i + 1


def literal(str_field: str):
    """
    converts string of object back to object
    example:
    $ str_literal('['a','b','c']')
    ['a', 'b', 'c'] # type is list
    """
    return ast.literal_eval(str_field)


def cumsum(collec, tuple_list_id=None, thresh=None):
    """
    Return index at which the cumulative sum of the collection exceeds the threshold.
    If the collection is a list of tuples, then a new list is created from only the
        indexed element of the tuple.
    """
    if tuple_list_id is not None:
        collec = [el[tuple_list_id] for el in collec]
    i = 0
    cumsum = 0
    if sum(collec) < thresh:
        print("threshold higher than sum of collection.")
        return len(collec)
    while cumsum < thresh:
        cumsum += collec[i]
        i += 1
    return i


def named_two_tuples(list_1, list_2, tuple_name='ntup', tuple_el_names='x y'):
    """
    Given two ordered lists of x and y
        return a list of namedtuples

    Example usage:
    >> named_two_tuples(feat_importance, feat_names,
                        tuple_name='fimp', tuple_el_names='importance name')
    """
    ntup = namedtuple(tuple_name, tuple_el_names)
    return [ntup(el[0], el[1]) for el in zip(list_1, list_2)]


def replace_rn(strobj):
    """
    There are characters such as "\r" and "\n" that perform line breaks (or returns),
    and these can create parsing dramas when reading/writing to disk.
    This helper function mitigates this by replacing it with a benign alternative.
    """
    return strobj.replace('\r', 'esc-r-').replace('\n', 'esc-n-')


def get_nested_item(data_dict, key_list):
    """
    obtain the deepest nested item in a nested dict given keys in a list.
    """
    item = data_dict.copy()
    for k in key_list:
        item = item[k]
    return item


def infix_to_postfix(infixexpr: str, token_regex: str, precedences: dict):
    """
    convert infix to postfix
    adapted from http://interactivepython.org/runestone/static/pythonds/BasicDS/InfixPrefixandPostfixExpressions.html

    Example usage:

    >> token_regex = "c[0-9]"
    >> prec = {"(": 1,
        "MINUS": 2,
        "OR": 3,
        "AND": 4}
    >> expr = '( c1 AND c2 ) MINUS c3'
    >> infix_to_postfix(expr, token_regex, prec)
    [out]:  'c1 c2 AND c3 MINUS'

    """
    op_stack = Stack()
    postfix_list = []
    token_list = infixexpr.split()
    for token in token_list:
        if re.search(token_regex, token):
            postfix_list.append(token)
        elif token == '(':
            op_stack.push(token)
        elif token == ')':
            top_token = op_stack.pop()
            while top_token != '(':
                postfix_list.append(top_token)
                top_token = op_stack.pop()
        else:
            while (not op_stack.isEmpty()) and \
               (precedences[op_stack.peek()] >= precedences[token]):
                  postfix_list.append(op_stack.pop())
            op_stack.push(token)

    while not op_stack.isEmpty():
        postfix_list.append(op_stack.pop())
    return " ".join(postfix_list)


def postfix_computer(tokens: list, logic: dict, operations: dict):
    """
    compute postfix expression given list of tokens.
     Each token in `tokens` must either be a key in `logic` or `operations`
     Example:
          op_dict = {'AND': lambda x,y: np.logical_and(x,y),
           'OR': lambda x,y: np.logical_or(x,y),
           'MINUS': lambda x,y: np.logical_and(x, np.logical_not(y))
          }
    """

    stack = []
    for token in tokens:
        if token in logic:
            stack.append(logic[token])
            continue

        op2, op1 = stack.pop(), stack.pop()
        if token in operations:
            stack.append(operations[token](op1, op2))

    return stack.pop()


def list_import_modules(library_root):
    """
    given the root dir of python library, return unique list of import modules.
      Useful for `install_requires` flag in setup.py, for example.
    """
    py_files = relevant_files(library_root, include_regex=".py", exclude=".pyc")
    import_lines = []
    for one_file in py_files:
        with open(one_file, "r") as f:
            lines = f.readlines()
        for i, line in enumerate(lines):
            if "import " in line:
                if "from" in line:
                    root = line.strip('\n').split('from ')[1]
                else:
                    root = line.strip('\n').split('import ')[1]
                root = root.split(' ')[0].split('.')[0]
                if root:
                    import_lines.append(root)
    return list(set(import_lines))
